/** @file
 *
 * This library contains functions for generating universally unique
 * identifiers.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
 */

/**
 * @ingroup Constants
 * @{
 */
private global const string _fwRPC_UUID_HEXCHARS = "0123456789abcdef";
private global const string _fwRPC_UUID_TEMPL_4 = "xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx";

private global const int _fwRPC_UUID_TEMPL_4_LEN = 36;
private global const int _fwRPC_UUID_RAND_DIVISOR = 32767 / 15;
/* @} */


/**
 * Generate a UUID v4 compliant string.
 * Uses internal WCCOA rand() function.
 *
 * @returns UUID v4 string
 *
 * @throws nothing
 */
string fwRPC_UUID_generateUUID4() {
  string result = "";
  char c;
  int random;

  for (int i=0; i<_fwRPC_UUID_TEMPL_4_LEN; ++i) {
    random = rand() / _fwRPC_UUID_RAND_DIVISOR;
    c = _fwRPC_UUID_TEMPL_4[i];

    switch(c) {
      case "x":
        result += _fwRPC_UUID_HEXCHARS[random];
        break;

      case "y":
        result += _fwRPC_UUID_HEXCHARS[(random & 0x3) + 8];
        break;

      default:
        result += c;
        break;
    }
  }

  return result;
}


/**
 * Check if a given string is a valid UUID v4 string.
 *
 * @param uuid (IN, string) UUID string to check
 *
 * @returns true  if uuid is a valid UUID v4 string
 *          false otherwise
 *
 * @throws nothing
 */
bool fwRPC_UUID_isValidUUID4(string uuid) {
  if (strlen(uuid) != _fwRPC_UUID_TEMPL_4_LEN) return false;

  char tmplChar;
  char cChar;

  for (int i=0; i<_fwRPC_UUID_TEMPL_4_LEN; ++i) {
    tmplChar = _fwRPC_UUID_TEMPL_4[i];
    cChar = uuid[i];

    switch(tmplChar) {
      case "x":
        if (cChar < "0" || cChar > "f") return false;
        break;

      case "y":
        if (cChar < "8" || cChar > "b") return false;
        break;

      default:
        if (cChar != tmplChar) return false;
        break;
    }
  }

  return true;
}
