/** @file
 *
 *  This library contains the RPC broker module functionality.
 *  Part of the server-side RPC library.
 *
 *  The broker module acts as a RB_MBX listener and is responsible for
 *  delegating messages to registered handler functions.
 *
 *  @author   Victor Rodrigues (CERN, BE-ICS-SDS)
 */

#uses "fwRPC/fwRPC_common.ctl"
#uses "fwRPC/fwRPC_RingBuffer.ctl"
#uses "fwRPC/fwRPC_UUID.ctl"
#uses "fwRPC/fwRPC_debug.ctl"


/**
 * Begin error codes
 */

global const int fwRPC_MSGBROKER_E_SUCCESS = 0;
global const int fwRPC_MSGBROKER_E_ERROR = -1;

/**
 * End error codes
 */


/**
 * @ingroup Constants
 * @{
 */
private global const int HDL_REG_SRC = 1;
private global const int HDL_REG_TARGET = 2;

private global const int MSG_DELEGATE_FUNC = 1;
private global const int MSG_DELEGATE_ACCEPTED = 2;
/* @} */


/**
 *  Mapping of registered message delegates.
 *  key   = message type
 *  value = DYN_MIXED:
 *            [MSG_DELEGATE_FUNC]
 *            [MSG_DELEGATE_ACCEPTED]: DYN_MIXED:
 *              [x]: client type
 */
private global mapping g_fwRPC_msgBroker_msgDelegates;

/**
 *  Mapping of registered handles, including source and destination info.
 *  key   = message handle
 *  value = dyn_mixed of dyn_string pairs of the following structure:
 *            fwRPC_DIST_NODE_PROJECT
 *            fwRPC_DIST_NODE_NAME
 *            HDL_REG_TARGET
 */
private global mapping g_fwRPC_msgBroker_hdl_registry;

/**
 * List of connected systems. Values are system numbers.
 */
private global dyn_uint g_fwRPC_msgBroker_connectedSystems;

/**
 *  Mapping of message buffers. Messages are buffered here if the destination
 *    ring buffer is full.
 *  key   = ringbuffer dp
 *  value = dyn_string messages
 */
private global mapping g_fwRPC_msgBroker_msgBuffers;


/**
 * Function to call when a new communication endpoint is detected
 */
private global string g_fwRPC_msgBroker_newMbxCB;

/**
 * Whether the broker is listening for local or remote endpoints
 */
private global bool g_fwRPC_msgBroker_localListen;

/**
 * Whether the broker is running or not
 */
private global bool g_fwRPC_msgBroker_isRunning = false;


//------------------------------------------------------------------------------


/**
 * Check if the broker is currently running or not.
 *
 * @returns true if running
 *          false otherwise
 *
 * @throws nothing
 */
bool fwRPC_msgBroker_isRunning() {
  return g_fwRPC_msgBroker_isRunning;
}


/**
 * Initialize the broker module. Resets local state.
 */
void fwRPC_msgBroker_initialize() {
  mappingClear(g_fwRPC_msgBroker_msgDelegates);
  mappingClear(g_fwRPC_msgBroker_hdl_registry);
  mappingClear(g_fwRPC_msgBroker_msgBuffers);
  dynClear(g_fwRPC_msgBroker_connectedSystems);
}


/**
 * Register a delegate function for a specific message type, and accept such
 *  requests only from specific source types.
 *
 * Requests inbound from unaccepted source types will not be sent to the
 *  delegate.
 *
 * Acceptable source types include
 *  fwRPC_DIST_TAG_CLIENT
 *  fwRPC_DIST_TAG_WORKER
 *  fwRPC_DIST_TAG_CORE
 *
 * Accepts a mapping of message type to delegate function names with the
 *  following signature:
 *
 *    void func(
 *      string project,
 *      string name,
 *      anytype buffer,
 *      string target_project,
 *      string target_name
 *    );
 *
 *  Where project is the requesting project
 *        name           is the requesting name
 *        buffer         is the deserialized message buffer
 *        target_project is the intended target project requested by the client
 *        target_name    is the intended target name requested by the client
 *
 *  target_project and target_name are optional parameters.
 *
 * @param msgType        (IN, string) message type
 * @param delegate       (IN, string) delegate function
 * @param acceptedSource (IN, string) accepted source type
 * @param ...            (IN, string) <accepted source types>
 *
 * @throws if the broker module is not running
 *         if the delegate was already registered
 */
void fwRPC_msgBroker_accept(
  string msgType,
  string delegate,
  string acceptedSource,
  ...
) {
  if (mappingHasKey(g_fwRPC_msgBroker_msgDelegates, msgType)) {
    fwException_throw(
      fwRPC_EX_INVALID_PARAM,
      "rule already exists for " + msgType
    );
  }

  dyn_string acceptedList = makeDynString(acceptedSource);

  va_list params;
  int len = va_start(params);
  for (int i=1; i<=len; ++i) {
    dynAppend(acceptedList, params[i]);
  }
  va_end(params);

  dyn_mixed entry;
  entry[MSG_DELEGATE_FUNC]     = delegate;
  entry[MSG_DELEGATE_ACCEPTED] = acceptedList;

  g_fwRPC_msgBroker_msgDelegates[msgType] = entry;
}


/**
 * Initialise the message broker and listen for outgoing messages from clients.
 *  The broker will listen for local or remote RB_MBX events as specified by the
 *  localListen parameter.
 *
 * The message broker itself does not create a RB_MBX DP.
 *
 *
 * Note that _stop() and _listen() should both be called from the same context,
 * i.e, within a monitor in order for dpDisconnect to succeed.
 *
 * @param localListen  (IN, bool)   whether to listen for local or remote RB_MBX
 *                                  events
 * @param newMbxCB     (IN, string) function to call when a new comm endpoint is
 *                                  detected
 *
 * @returns fwRPC_MSGBROKER_E_SUCCESS on success
 *          fwRPC_MSGBROKER_E_ERROR   otherwise
 *
 * @throws if the parameters are invalid
 *         if the broker is already running
 */
int fwRPC_msgBroker_listen(bool localListen, string newMbxCB = "") {
  _fwRPC_msgBroker_assertRunning(false);

  //This must be set before any external CBs are called in case an attempt is
  //made to use the broker manager before this method is complete.
  g_fwRPC_msgBroker_isRunning = true;

  if (!mappinglen(g_fwRPC_msgBroker_msgDelegates)) {
    fwRPC_dbg_log_module(
      "msg broker", fwRPC_DBG_EVENT_WARNING,
      "listen: WARNING: delegates list is empty before listen, messages may" +
      " be skipped!"
    );
  }

  g_fwRPC_msgBroker_newMbxCB = newMbxCB;
  g_fwRPC_msgBroker_localListen = localListen;

  int err = _fwRPC_msgBroker_listenSystem(localListen, newMbxCB);
  if (err < 0) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "listen: error connecting to local RB_MBX"
    );
    g_fwRPC_msgBroker_isRunning = false;
    return fwRPC_MSGBROKER_E_ERROR;
  }

  if (!localListen) {
    err = dpConnectUserData(
      "_fwRPC_msgBroker_newSystemCB",
      newMbxCB,
      true,
      "_DistManager.State.LastUpdateTime"
    );

    if (err < 0 || dynlen(getLastError())) {
      fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
        "listen: Error connecting to dist manager. New systems will not be" +
        " discovered."
      );
    }
  }

  return fwRPC_MSGBROKER_E_SUCCESS;
}


/**
 * Stop the currently running message broker.
 *
 * Note that _stop() and _listen() should both be called from the same context,
 * i.e, within a monitor in order for dpDisconnect to succeed.
 *
 * @returns fwRPC_MSGBROKER_E_SUCCESS on success
 *          fwRPC_MSGBROKER_E_ERROR   on error
 *
 * @throws if the broker is not running
 */
int fwRPC_msgBroker_stop() {
  _fwRPC_msgBroker_assertRunning(true);

  int err;

  fwRPC_msgBroker_initialize();

  err = dpQueryDisconnect("_fwRPC_msgBroker_msgInCB", 0);
  if (err < 0 || dynlen(getLastError())) return fwRPC_MSGBROKER_E_ERROR;

  err = dpQueryDisconnect("_fwRPC_msgBroker_rbDrainCB", 0);
  if (err < 0 || dynlen(getLastError())) return fwRPC_MSGBROKER_E_ERROR;

  if (g_fwRPC_msgBroker_newMbxCB != "") {
    err = dpQueryDisconnect(
      "_fwRPC_msgBroker_newMbxCB",
      g_fwRPC_msgBroker_newMbxCB
    );
    if (err < 0 || dynlen(getLastError())) return fwRPC_MSGBROKER_E_ERROR;
  }

  if (!g_fwRPC_msgBroker_localListen) {
    err = dpDisconnectUserData(
      "_fwRPC_msgBroker_newSystemCB",
      g_fwRPC_msgBroker_newMbxCB,
      "_DistManager.State.LastUpdateTime"
    );
    if (err < 0 || dynlen(getLastError())) return fwRPC_MSGBROKER_E_ERROR;
  }

  g_fwRPC_msgBroker_isRunning = false;

  return fwRPC_MSGBROKER_E_SUCCESS;
}


/**
 * Write a message directly to the destination RB_MBX input.
 *
 * @param targetProject (IN, string) destination project
 * @param targetName    (IN, string) destination name
 * @param type          (IN, string) message type
 * @param payload       (IN, mixed)  message contents
 *
 * @returns fwRPC_MSGBROKER_E_SUCCESS on success
 *          fwRPC_MSGBROKER_E_ERROR   otherwise
 *
 * @throws if the broker module is not running
 */
synchronized int fwRPC_msgBroker_writeMessage(
  string targetProject, string targetName,
  string type, mixed payload = "",
  string sourceProject="", string sourceName=""
) {
  _fwRPC_msgBroker_assertRunning(true);

  string message;
  string ringbufferDp;

  if (sourceProject == "") sourceProject = fwRPC_DIST_BRKR;
  if (sourceName == "") sourceName = fwRPC_DIST_BRKR;

  message = fwRPC_dist_buildMessage(
    type, 
    sourceProject,
    sourceName,
    payload
  );

  ringbufferDp = targetProject + ":" +
                 fwRPC_common_makeDpName(fwRPC_RBMBX_DPT, targetName) +
                 ".in";

  if (mappingHasKey(g_fwRPC_msgBroker_msgBuffers, ringbufferDp)) {
    dynAppend(g_fwRPC_msgBroker_msgBuffers[ringbufferDp], message);
    return fwRPC_MSGBROKER_E_SUCCESS;
  }

  int err = fwRPC_RingBuffer_write(ringbufferDp, message);
  if (err == fwRPC_RingBuffer_EFULL) {
    g_fwRPC_msgBroker_msgBuffers[ringbufferDp] = makeDynString(message);
  } else if (err != fwRPC_RingBuffer_ESUCCESS) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
    "writeMessage: Error writing to buffer: " + ringbufferDp + " (" + err + ")"
    );

    return fwRPC_MSGBROKER_E_ERROR;
  }

  return fwRPC_MSGBROKER_E_SUCCESS;
}


/**
 * Retrieve the callee of a registered handle.
 *
 * @param hdl     (IN, anytype)  handle to check for
 * @param project (OUT, string)  handle callee project
 * @param name    (OUT, string)  handle callee name
 *
 * @returns fwRPC_MSGBROKER_E_SUCCESS  on success
 *          fwRPC_MSGBROKER_E_ERROR    otherwise, or if the handle had not been 
 *                                     registered
 *
 * @throws if the broker module is not running
 */
int fwRPC_msgBroker_getHandleCallee(
  anytype hdl,
  string &project, string &name
) {
  _fwRPC_msgBroker_assertRunning(true);

  if (!mappingHasKey(g_fwRPC_msgBroker_hdl_registry, hdl)) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "getCallee: Function was never called: " + hdl
    );
    return fwRPC_MSGBROKER_E_ERROR;
  }

  dyn_mixed handleData = g_fwRPC_msgBroker_hdl_registry[hdl];

  //check if the handle was registered with a callee value or not
  if (dynlen(handleData) < HDL_REG_TARGET) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "getCallee: call was never attributed a target " + hdl
    );
    return fwRPC_MSGBROKER_E_ERROR;
  }

  dyn_string target = handleData[HDL_REG_TARGET];
  project = target[fwRPC_DIST_NODE_PROJECT];
  name    = target[fwRPC_DIST_NODE_NAME];

  return fwRPC_MSGBROKER_E_SUCCESS;
}


/**
 * Retrieve the caller of a previously registered handle.
 *
 * @param hdl     (IN, anytype)  handle to check for
 * @param project (OUT, string) handle caller project
 * @param name    (OUT, string) handle caller name
 *
 * @returns fwRPC_MSGBROKER_E_SUCCESS on success
 *          fwRPC_MSGBROKER_E_ERROR   otherwise
 *
 * @throws if the broker module is not running
 */
int fwRPC_msgBroker_getHandleCaller(
  anytype hdl,
  string &project, string &name
) {
  _fwRPC_msgBroker_assertRunning(true);

  if (!mappingHasKey(g_fwRPC_msgBroker_hdl_registry, hdl)) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "getCaller: Function was never called: " + hdl
    );
    return fwRPC_MSGBROKER_E_ERROR;
  }

  dyn_string source = g_fwRPC_msgBroker_hdl_registry[hdl][HDL_REG_SRC];
  project = source[fwRPC_DIST_NODE_PROJECT];
  name    = source[fwRPC_DIST_NODE_NAME];

  return fwRPC_MSGBROKER_E_SUCCESS;
}


/**
 * Register a caller and callee with a specific handle. If the handle is already
 * registered, no insertion is performed.
 *
 * Once registered, the callee of the handle can be specified using the
 * _setHandleCallee() function.
 *
 * @param hdl           (IN, anytype) handle to register
 * @param senderProject (IN, string)  caller project
 * @param senderName    (IN, string)  caller name
 * @param targetProject (IN, string)  callee project
 * @param targetName    (IN, string)  callee name
 *
 * @returns fwRPC_MSGBROKER_E_SUCCESS on success
 *          fwRPC_MSGBROKER_E_ERROR   if the entry already exists
 *
 * @throws if the broker module is not running
 */
int fwRPC_msgBroker_registerHandle(
  anytype hdl,
  string senderProject, string senderName,
  string targetProject="", string targetName=""
) synchronized(g_fwRPC_msgBroker_hdl_registry) {
  _fwRPC_msgBroker_assertRunning(true);

  if (mappingHasKey(g_fwRPC_msgBroker_hdl_registry, hdl)) {
    return fwRPC_MSGBROKER_E_ERROR;
  };

  dyn_mixed call;
  dyn_string pair;

  pair[fwRPC_DIST_NODE_PROJECT] = senderProject;
  pair[fwRPC_DIST_NODE_NAME]    = senderName;
  call[HDL_REG_SRC] = pair;

  if (targetProject != "" && targetName != "") {
    pair[fwRPC_DIST_NODE_PROJECT] = targetProject;
    pair[fwRPC_DIST_NODE_NAME]    = targetName;
    call[HDL_REG_TARGET] = pair;
  }

  g_fwRPC_msgBroker_hdl_registry[hdl] = call;

  return fwRPC_MSGBROKER_E_SUCCESS;
}


/**
 * Set the target (callee) of a previously registered handle.
 *
 * Handles must be registered beforehand using the _registerHandle() function.
 *
 * @param hdl     (IN, anytype) handle to update
 * @param project (IN, string)  callee project
 * @param name    (IN, string)  callee name
 *
 * @returns fwRPC_MSGBROKER_E_SUCCESS on success
 *          fwRPC_MSGBROKER_E_ERROR   otherwise, or if the handle had not been 
 *                                    registered
 *
 * @throws if the broker module is not running
 */
int fwRPC_msgBroker_setHandleCallee(
  anytype hdl,
  string project,
  string name
) synchronized(g_fwRPC_msgBroker_hdl_registry) {
  _fwRPC_msgBroker_assertRunning(true);

  if (!mappingHasKey(g_fwRPC_msgBroker_hdl_registry, hdl)) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "setTarget: call not registered: " + hdl
    );
    return fwRPC_MSGBROKER_E_ERROR;
  }

  dyn_string pair;
  pair[fwRPC_DIST_NODE_PROJECT] = project;
  pair[fwRPC_DIST_NODE_NAME] = name;
  g_fwRPC_msgBroker_hdl_registry[hdl][HDL_REG_TARGET] = pair;

  return fwRPC_MSGBROKER_E_SUCCESS;
}


/**
 * Remove a handle from the registry.
 *
 * @param hdl (IN, anytype) handle to remove
 *
 * @returns fwRPC_MGSBROKER_E_SUCCESS on success
 *          fwRPC_MSGBROKER_E_ERROR   if the handle had not been registered
 *
 * @throws if the broker module is not running
 */
int fwRPC_msgBroker_unregisterHandle(anytype hdl)
synchronized(g_fwRPC_msgBroker_hdl_registry) {
  _fwRPC_msgBroker_assertRunning(true);

  if (!mappingHasKey(g_fwRPC_msgBroker_hdl_registry, hdl)) {
    return fwRPC_MSGBROKER_E_ERROR;
  }

  mappingRemove(g_fwRPC_msgBroker_hdl_registry, hdl);

  return fwRPC_MSGBROKER_E_SUCCESS;
}


/**
 * Retrieve a list of all handles registered with a given caller.
 *
 * @param project (IN, string)      caller project
 * @param name    (IN, string)      caller name
 * @param hdls    (OUT, dyn_string) list of handles
 *
 * @returns nothing
 *
 * @throws if the broker module is not running
 */
void fwRPC_msgBroker_getHandlesCalledBy(
  string project,
  string name,
  dyn_string &hdls
) synchronized(g_fwRPC_msgBroker_hdl_registry) {
  _fwRPC_msgBroker_assertRunning(true);

  hdls = makeDynString();

  int regsToCheck = mappinglen(g_fwRPC_msgBroker_hdl_registry);
  for (int i = 1; i <= regsToCheck; ++i) {
    dyn_dyn_string call = mappingGetValue(g_fwRPC_msgBroker_hdl_registry, i);
    dyn_string src = call[HDL_REG_SRC];

    if (
      src[fwRPC_DIST_NODE_PROJECT] == project &&
      src[fwRPC_DIST_NODE_NAME] == name
    ) {
      dynAppend(hdls, mappingGetKey(g_fwRPC_msgBroker_hdl_registry, i));
    }
  }
}


//------------------------------------------------------------------------------


/**
 * @internal
 *
 * New system connection callback. Tracks the currently connected systems and
 *  automatically query (dis)connects to the remote RB_MBXs.
 *
 * @param newMbxCB       (IN, string) callback to execute on detection of new mbx
 * @param dpe            (IN, string) triggering DPE (ignored)
 * @param lastUpdateTime (IN, time)   new system detection time (ignored)
 *
 * @returns nothing
 *
 * @throws nothing
 */
synchronized private void _fwRPC_msgBroker_newSystemCB(
  string newMbxCB,
  string dpe,
  time lastUpdateTime
) {
  dyn_string systemNames;
  dyn_uint systemIds;
  int err;
  string localSystemName;

  getSystemNames(systemNames, systemIds);
  localSystemName = fwRPC_common_cleanProjectName(getSystemName());

  //connect to new systems
  int systemsToCheck = dynlen(systemNames);
  for (int i = 1; i <= systemsToCheck; ++i) {
    if (dynContains(g_fwRPC_msgBroker_connectedSystems, systemIds[i])) continue;
    if (systemNames[i] == localSystemName) continue;

    dynAppend(g_fwRPC_msgBroker_connectedSystems, systemIds[i]);

    err = _fwRPC_msgBroker_listenSystem(false, newMbxCB, systemNames[i]);
    if (err < 0) {
      fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
        "new system cb: Error connecting to RB_MBX on " + systemNames[i]
      );
    }
  }

  //disconnect from the systems we've lost
  int connectedSystems = dynlen(g_fwRPC_msgBroker_connectedSystems);
  for (int i = 1; i <= connectedSystems; ++i) {
    if (dynContains(systemIds, g_fwRPC_msgBroker_connectedSystems[i])) continue;

    err = dpQueryDisconnect("_fwRPC_msgBroker_msgInCB", 0);
    if (err < 0 || dynlen(getLastError())) {
      fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
        "new system cb: dp query disconnect error: " + err
      );
    }

    err = dpQueryDisconnect("_fwRPC_msgBroker_newMbxCB", newMbxCB);
    if (err < 0 || dynlen(getLastError())) {
      fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
        "new system cb: dp query disconnect error: " + err
      );
    }

    dynRemove(g_fwRPC_msgBroker_connectedSystems, i);
  }
}


/**
 * @internal
 *
 * Retrieve client information from dpQuery dpe result on an RBMBX.
 *
 * @param dpe         (IN,  string) dpe
 * @param dp          (OUT, string) ringbuffer MBX dp
 * @param systemName  (OUT, string) name of system
 * @param name        (OUT, string) name of project
 *
 * @returns void
 *
 * @throws nothing
 */
private void _fwRPC_msgBroker_getProjectInfoFromDpe(
  string dpe,
  string &dp,
  string &systemName,
  string &name
) {
  string project = dpSubStr(dpe, DPSUB_SYS);
  dp = dpSubStr(dpe, DPSUB_DP);
  systemName = fwRPC_common_cleanProjectName(project);
  name = substr(dp, strlen(fwRPC_RBMBX_DPT) + 1);
}


/**
 * @internal
 *
 * Remote RB_MBX initialization callback.
 *
 * @param cb      (IN, string) callback to execute upon detection of new mbx
 * @param results (IN, dyn_dyn_anytype) query result
 *
 * @returns nothing
 *
 * @throws nothing
 */
synchronized private void _fwRPC_msgBroker_newMbxCB(
  string cb, dyn_dyn_anytype results
) {
  for (int i=2; i<=dynlen(results); ++i) {
    if (!results[i][2]) continue;
    string dp, project, name;
    _fwRPC_msgBroker_getProjectInfoFromDpe(results[i][1], dp, project, name);

    startThread(cb, project, name);
  }
}


/**
 * @internal
 *
 * Remote outbound message handler. Callback to dpQueryConnectSingle.
 *
 * @TODO: this needs clearing up..
 *
 * @param _       (IN, anytype) ignored
 * @param results (IN, dyn_dyn_anytype) query results
 *
 * @returns nothing
 *
 * @throws nothing
 */
synchronized private void _fwRPC_msgBroker_msgInCB(
  anytype _,
  dyn_dyn_anytype results
) {
  for (int i=2; i<=dynlen(results); ++i) {
    string dp, senderProject, senderName;
    _fwRPC_msgBroker_getProjectInfoFromDpe(
      results[i][1],
      dp,
      senderProject,
      senderName
    );

    string ringbufferDp = senderProject + ":" + dp + ".out";

    string val;
    int err;
    while ((err = fwRPC_RingBuffer_read(ringbufferDp, val)) != fwRPC_RingBuffer_EEMPTY) {
      if (val == "") continue;

      string msgId, msgType, targetProject, targetName, buffer;

      fwRPC_common_unpackMessage(
        val,
        msgId,
        msgType,
        targetProject,
        targetName,
        buffer
      );

      if (!mappingHasKey(g_fwRPC_msgBroker_msgDelegates, msgType)) {
        fwRPC_dbg_log(
          fwRPC_DBG_EVENT_WARNING,
          "delegate not registered for msg type of " + msgType
        );
        continue;
      }

      anytype deserialized;
      deserializeFromXml(buffer, deserialized);

      dyn_mixed delegate = g_fwRPC_msgBroker_msgDelegates[msgType];

      bool valid = false;
      dyn_string acceptedNames = delegate[MSG_DELEGATE_ACCEPTED];
      int namesToCheck = dynlen(acceptedNames);
      for (int j = 1; j <= namesToCheck; ++j) {
        if (fwRPC_dist_nameMatches(senderName, acceptedNames[j])) {
          valid = true;
          break;
        }
      }

      if (!valid) {
        fwRPC_dbg_log(
          fwRPC_DBG_EVENT_WARNING,
          "messages not accepted from " + senderName + " for type " + msgType
        );
        continue;
      }

      string func = delegate[MSG_DELEGATE_FUNC];
      startThread(
        func,
        senderProject, senderName,
        deserialized,
        targetProject, targetName
      );
    }

    if (err != fwRPC_RingBuffer_EEMPTY) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        "Error reading from ringbuffer: " + err
      );
    }
  }
}


/**
 * @internal
 *
 * Ringbuffer drain event CB. Evokes queue processing.
 * Called when a monitored ringbuffer has been read from, indicating that we can
 * write to it again in the case of a previously full buffer.
 *
 * @param _       (IN, anytype) ignored
 * @param results (IN, dyn_dyn_string) dpQuery results
 *
 * @returns nothing
 *
 * @throws nothing
 */
synchronized private void _fwRPC_msgBroker_rbDrainCB(
  anytype _,
  dyn_dyn_string results
) {
  for (int i=2; i<=dynlen(results); ++i) {
    //trim the 'read_idx' part..
    string dpe = results[i][1];
    string rbDp = substr(dpe, 0, strlen(dpe) - strlen(fwRPC_RingBuffer_DPE_READ));
    if (!mappingHasKey(g_fwRPC_msgBroker_msgBuffers, rbDp)) {
      //buffer doesn't exist, equivalent to empty..
      return;
    }

    _fwRPC_msgBroker_processOverflowBuffer(rbDp);
  }
}


/**
 * @internal
 *
 * Process the message queue and attempt to write to the destination ring
 * buffer.
 *
 * @param dp (IN, string) dp of the target ring buffer
 *
 * @returns nothing
 * 
 * @throws nothing
 */
private void _fwRPC_msgBroker_processOverflowBuffer(string dp) {
  dyn_string queue = g_fwRPC_msgBroker_msgBuffers[dp];

  while(dynlen(queue) > 0) {
    string msg = queue[1];

    int err = fwRPC_RingBuffer_write(dp, msg);
    if (err < 0) {
      g_fwRPC_msgBroker_msgBuffers[dp] = queue;
      return;
    }

    dynRemove(queue, 1);
  }

  mappingRemove(g_fwRPC_msgBroker_msgBuffers, dp);
}


/**
 * @internal
 *
 * Build a query for read/write pointer changes in a ringbuffer DP in a
 * RB_MBX DP.
 *
 * Produces a DPE name resembling the following:
 *  .in.write_idx
 *
 * @param read   (IN, bool)   listen for read or write iterator changes
 *
 * @returns the resulting query string
 *
 * @throws nothing
 */
private string _fwRPC_msgBroker_buildRbPointerDpe(
  bool read
) {
  string dp_sub = read ? ".in" : ".out";
  string dp_itr = read ? fwRPC_RingBuffer_DPE_READ : fwRPC_RingBuffer_DPE_WRITE;

  return dp_sub + dp_itr;
}


/**
 * @internal
 *
 * Build a query for arbitrary DPE on RB_MBX DPs.
 *
 * @param local  (IN, bool)   listen for local or remote RB_MBX
 * @param rb_dpe (IN, string) RingBuffer element to listen for
 * @param remote (IN, string) system to listen on
 *
 * @returns the resulting query string
 *
 * @throws nothing
 */
private string _fwRPC_msgBroker_buildRbMbxQuery(
  bool local,
  string rb_dpe,
  string remote = ""
) {
  string locality = local ? fwRPC_DIST_TAG_LOCAL : fwRPC_DIST_TAG_REMOTE;
  string full_dpe = fwRPC_RBMBX_DPT + fwRPC_DP_SEPARATOR +
                    locality + "_*" + rb_dpe;

  //full_dpe should look something like: _FwRPC_RingBuffer/loc_*.in.write_idx
  
  string query =  "SELECT '_online.._value'" +
                  " FROM '{" + full_dpe + "}'" +
                  (remote ? " REMOTE '" + remote + ":'" : "") +
                  " WHERE _DPT =\"" + fwRPC_RBMBX_DPT + "\"";

  return query;
}


/**
 * @internal
 *
 * Listen for changes in RB_MBX in a given system. Registers msg input and drain
 * event callback functions. Also registers new mbx event callback if one is
 * specified using the newMbxCB parameter.
 *
 * If specified, newMbxCB should respect the following signature:
 *  void newMbxCB(string project, string name)
 * Where project and name will refernce the new mbx.
 *
 * @param localListen (IN, bool)   listen for local or remote RB_MBX
 * @param newMbxCB    (IN, string) CB for new RB_MBX events
 * @param systemName  (IN, string) system to listen for
 *
 * @returns fwRPC_MSGBROKER_E_SUCCESS on success
 *          fwRPC_MSGBROKER_E_ERROR   otherwise
 *
 * @throws nothing
 */
private int _fwRPC_msgBroker_listenSystem(
  bool localListen,
  string newMbxCB = "",
  string systemName = ""
) {
  int err;

  string dpe = _fwRPC_msgBroker_buildRbPointerDpe(false);
  string outputRbQuery = _fwRPC_msgBroker_buildRbMbxQuery(
    localListen,
    dpe,
    systemName
  );

  err = dpQueryConnectSingle("_fwRPC_msgBroker_msgInCB", true, 0, outputRbQuery);
  if (err < 0 || dynlen(getLastError())) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "listen system: dp query connect error: " + err
    );
    return fwRPC_MSGBROKER_E_ERROR;
  }

  dpe = _fwRPC_msgBroker_buildRbPointerDpe(true);
  string inputRbQuery = _fwRPC_msgBroker_buildRbMbxQuery(
    localListen,
    dpe,
    systemName
  );

  err = dpQueryConnectSingle("_fwRPC_msgBroker_rbDrainCB", false, 0, inputRbQuery);
  if (err < 0 || dynlen(getLastError())) {
    fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
      "listen system: dp query connect error: " + err
    );

    dpQueryDisconnect("_fwRPC_msgBroker_msgInCB", 0);
    return fwRPC_MSGBROKER_E_SUCCESS;
  }

  if (newMbxCB != "") {
    string rbTriggerQuery = _fwRPC_msgBroker_buildRbMbxQuery(
      localListen,
      ".initTrigger",
      systemName
    );

    err = dpQueryConnectSingle(
      "_fwRPC_msgBroker_newMbxCB",
      true,
      newMbxCB,
      rbTriggerQuery
    );

    if (err < 0 || dynlen(getLastError())) {
      fwRPC_dbg_log(fwRPC_DBG_EVENT_ERROR,
        "listen system: dp query connect error: " + err
      );
      dpQueryDisconnect("_fwRPC_msgBroker_msgInCB", 0);
      dpQueryDisconnect("_fwRPC_msgBroker_rbDrainCB", newMbxCB);
      return fwRPC_MSGBROKE_E_ERROR;
    }
  }

  return fwRPC_MSGBROKER_E_SUCCESS;
}


/**
 * @internal
 *
 * Assert that the broker is in the specified running state.
 *
 * @param running (IN, bool) assert whether running or not running
 *
 * @returns nothing
 *
 * @throws if the running state does not match
 */
private void _fwRPC_msgBroker_assertRunning(bool running) {
  if (g_fwRPC_msgBroker_isRunning != running) {
    fwException_throw(
      fwRPC_EX_BROKER_STATE,
      "Broker should " + (running ? "" : "not") + " be running."
    );
  }
}
