/** @file
 *
 * This library contains constants and functionality for workers, including DP,
 *  query, and model information. Also includes functionality for worker model
 *  initialization and synchronization.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
 */

#uses "fwRPC/fwRPC_common.ctl"

#uses "fwRPC/fwRPC_exception.ctl"

/**
 * Maximum number of jobs to be executed concurrently by a single worker.
 * @ingroup Constants
 */
const int fwRPC_WORKER_MAX_JOBS = 25;

/**
 * Worker heartbeat parameters
 * @ingroup Constants
 * @{
 */
const int fwRPC_WORKER_HEARTBEAT_WAIT_S = 5;
const int fwRPC_WORKER_HEARTBEAT_RATE_MS = 2000;
/* @} */


/**
 * States in which a worker can be in
 * @ingroup Constants
 * @{
 */
const int fwRPC_WORKER_STATUS_NONE       = 1;
const int fwRPC_WORKER_STATUS_READY      = 2;
const int fwRPC_WORKER_STATUS_PENDINGRUN = 3;
const int fwRPC_WORKER_STATUS_WORKING    = 4;
const int fwRPC_WORKER_STATUS_ERROR      = 5;
const int fwRPC_WORKER_STATUS_DEAD       = 6;
const int fwRPC_WORKER_STATUS_RESTARTING = 7;
const int fwRPC_WORKER_STATUS_MAXIDX     = 8;
/* @} */


/**
 * Worker model indexes
 *
 * @ingroup Constants
 * @{
 */
const int fwRPC_WORKER_MODEL_ID     = 1;
const int fwRPC_WORKER_MODEL_STATUS = 2;
const int fwRPC_WORKER_MODEL_JOBS   = 3;
const int fwRPC_WORKER_MODEL_LOCKED = 4;
const int fwRPC_WORKER_MODEL_LASTHB = 5;
const int fwRPC_WORKER_MODEL_HBTID  = 6;
const int fwRPC_WORKER_MODEL_PEND_RESTART = 7;
const int fwRPC_WORKER_MODEL_MAXIDX = 8;
/* @} */


//------------------------------------------------------------------------------


/**
 * Construct a worker object.
 *
 * @param id     (IN, int)  Worker ID
 * @param status (IN, int)  Initial status
 * @param jobs   (IN, int)  Initial job count
 * @param locked (IN, bool) Initial lock status
 *
 * @returns Worker object
 *
 * @throws nothing
 */
dyn_mixed fwRPC_workerModel_new(
  int id,
  int status = fwRPC_WORKER_STATUS_NONE,
  int jobs = 0,
  bool locked = false
) {
  dyn_mixed worker;

  worker[fwRPC_WORKER_MODEL_ID] = id;
  worker[fwRPC_WORKER_MODEL_STATUS] = status;
  worker[fwRPC_WORKER_MODEL_JOBS] = jobs;
  worker[fwRPC_WORKER_MODEL_LOCKED] = locked;
  worker[fwRPC_WORKER_MODEL_PEND_RESTART] = false;
  worker[fwRPC_WORKER_MODEL_MAXIDX] = 0;

  return worker;
}


/**
 * Check if the given object is a valid worker object. Throws an exception if
 *   not.
 *
 * @param worker (IN, dyn_mixed) object to check
 *
 * @returns nothing
 *
 * @throws if the given object is not a valid worker object
 */
bool fwRPC_workerModel_isValid(dyn_mixed worker) {
  if (dynlen(worker) < fwRPC_WORKER_MODEL_MAXIDX) return false;
  return true;
}


/**
 * Return whether or not a given worker is available for work given it's status
 *  (e.g., if it is not currently locked in execution), and the exclusive
 *  execution requirements of the request, if any.
 *
 * @param worker        (IN, dyn_mixed) worker to check
 * @param exclusiveExec (IN, bool)      whether the request required exclusive
 *                                      execution.
 *
 * @returns true  if the worker is suitable
 *          false if the worker is not suitable
 *
 * @throws nothing
 */
bool fwRPC_workerModel_canProcess(dyn_mixed worker, bool funcIsExclusive) {
  if (!fwRPC_workerModel_isValid(worker)) return false;

  int jobCount = worker[fwRPC_WORKER_MODEL_JOBS];
  int status = worker[fwRPC_WORKER_MODEL_STATUS];

//  if (worker[fwRPC_WORKER_MODEL_PEND_RESTART])
//    return false;

  if (worker[fwRPC_WORKER_MODEL_LOCKED])
    return false;

  if (funcIsExclusive && (status != fwRPC_WORKER_STATUS_READY || jobCount > 0)) 
    return false;

  if (status == fwRPC_WORKER_STATUS_ERROR || status == fwRPC_WORKER_STATUS_DEAD)
    return false;

  if (jobCount >= fwRPC_WORKER_MAX_JOBS)
    return false;

  return true;
}


/**
 * Return whether or not the worker is idle. Note that whether the worker is
 * pending a restart or not is ignored in this evaluation.
 *
 * @param worker (IN, dyn_mixed) worker object
 *
 * @returns true  if idle
 *          false otherwise or if worker is invalid
 *
 * @throws nothing
 */
bool fwRPC_workerModel_isIdle(dyn_mixed worker) {
  if (!fwRPC_workerModel_isValid(worker)) return false;

  if (worker[fwRPC_WORKER_MODEL_JOBS]) return false;
  if (worker[fwRPC_WORKER_MODEL_STATUS] != fwRPC_WORKER_STATUS_READY) return false;
  if (worker[fwRPC_WORKER_MODEL_LOCKED]) return false;

  return true;
}


string fwRPC_workerModel_toString(dyn_mixed worker) {
  int status = worker[fwRPC_WORKER_MODEL_STATUS];

  string str;
  str  = "id:      " + worker[fwRPC_WORKER_MODEL_ID] + " ";
  str += "status:  " + fwRPC_workerModel_status_toString(status) + " ";
  str += "jobs:    " + worker[fwRPC_WORKER_MODEL_JOBS] + " ";
  str += "locked:  " + worker[fwRPC_WORKER_MODEL_LOCKED] + " ";
  str += "lasthb:  " + (string)worker[fwRPC_WORKER_MODEL_LASTHB] + " ";
  str += "hbtid:   " + worker[fwRPC_WORKER_MODEL_HBTID] + " ";
  str += "restart: " + worker[fwRPC_WORKER_MODEL_PEND_RESTART] + " ";

  return str;
}

string fwRPC_workerModel_status_toString(int status) {
  if (status < 1 || status >= fwRPC_WORKER_STATUS_MAXIDX) return "";

  dyn_string strings;
  strings[fwRPC_WORKER_STATUS_NONE]       = "none";
  strings[fwRPC_WORKER_STATUS_READY]      = "ready";
  strings[fwRPC_WORKER_STATUS_PENDINGRUN] = "pending run";
  strings[fwRPC_WORKER_STATUS_WORKING]    = "working";
  strings[fwRPC_WORKER_STATUS_ERROR]      = "error";
  strings[fwRPC_WORKER_STATUS_DEAD]       = "dead";
  strings[fwRPC_WORKER_STATUS_RESTARTING] = "restarting";

  return strings[status];
}
