/** @file
 *
 * This file contains commonly used constants and functions for unit testing.
 *
 * @author victor Rodrigues (CERN, BE-ICS-SDS)
 */

#uses "fwRPC/fwRPC_client.ctl"
#uses "fwRPC/fwRPC_common.ctl"
#uses "fwRPC/fwRPC_RingBuffer.ctl"

#uses "fwRPC/fwRPC_debug.ctl"

private global const bool _PRINT_SUITE = true;
private global const bool _PRINT_TEST  = true;
private global const bool _PRINT_EMPTY = true;
private global int _TEST_COUNT;


//SETUP UTILS
void stopWorkers(dyn_mixed coninfo) {
  dyn_dyn_mixed managers;

  fwRPC_pmon_getWorkerCtrlManagers(coninfo, managers);

  for (int i=1; i<=dynlen(managers); ++i) {
    dyn_mixed mngr = managers[i];
    if (mngr[fwRPC_PMON_MANINFO_STATE] != fwRPC_PMON_MANINFO_STATE_STOPPED) {
      fwRPC_pmon_stopCtrlManager(coninfo, mngr[fwRPC_PMON_MANINFO_INDEX]);
    }
  }
}


void clearFuncDp(string name) {
  string dp = fwRPC_common_makeDpName(fwRPC_FUNCTION_DPT, name);
  if (dpExists(dp)) dpDelete(dp);
}


//------------------------------------------------------------------------------


private global int _maxJobs;
void _startJobWatch(int maxJobs, string dpWorkerRb) {
  _maxJobs = maxJobs;

  int err = dpConnectUserData("_jobwatchCB", maxJobs, false, dpWorkerRb + ".out.values");
  assert(err == 0 && dynlen(getLastError()) == 0, "Error in DP connect");
}

void _stopJobWatch(string dpWorkerRb) {
  int err = dpDisconnectUserData("_jobwatchCB", _maxJobs, dpWorkerRb + ".out.values");
  assert(err == 0 && dynlen(getLastError()) == 0, "Error in DP disconnect");
}

synchronized void _jobwatchCB(int max, string dpe, dyn_string messages) {
  for (int i=1; i<=dynlen(messages); ++i) {
    string id, type, proj, name, buffer;
    fwRPC_common_unpackMessage(messages[i], id,type,proj,name,buffer);

    if (type != fwRPC_DISTMSG_TYPE_WORKER_STATE) continue;

    anytype deserialized;
    deserializeFromXml(buffer, deserialized);

    int jobs = deserialized[fwRPC_DISTMSG_WORKER_STATE_JOBS];
    assert(jobs <= max, "jobs > " + max + ": " + jobs);
  }
}


//------------------------------------------------------------------------------
// request aides
//------------------------------------------------------------------------------


private global mapping _asyncStatusMap;

int _waitForCompletion(
  anytype ctx,
  anytype hdl,
  int     &status,
  anytype &result,
  int     timeout = 10000
) {
  fwRPC_client_subscribeCallStatus(ctx, hdl, "mock_statusCB");

  int inc = 500;
  int _status;

  do {
    _status = _getLastStatus(hdl);
    delay(0, inc);
    timeout -= inc;
  } while (
    timeout > 0 &&
    _status != fwRPC_REQUEST_STATUS_DONE &&
    _status != fwRPC_REQUEST_STATUS_ERROR &&
    _status != fwRPC_REQUEST_STATUS_CANCEL
  );

  assert(timeout > 0, "call timeout reached, last status: " + _status + " for " + hdl);

  status = _status;
  return fwRPC_client_getResult(ctx, hdl, result);
}


void _subscribeHdl(anytype ctx, anytype hdl) {
  fwRPC_client_subscribeCallStatus(ctx, hdl, "mock_statusCB");
}


int _getLastStatus(anytype hdl) {
  return mappingHasKey(_asyncStatusMap, hdl) ?
          _asyncStatusMap[hdl][dynlen(_asyncStatusMap[hdl])] :
          fwRPC_REQUEST_STATUS_NONE;
}

dyn_int _getAllStatuses(anytype hdl) {
  return mappingHasKey(_asyncStatusMap, hdl) ?
          _asyncStatusMap[hdl] : makeDynInt();
}

void _clearStatuses(anytype hdl) {
  mappingRemove(_asyncStatusMap, hdl);
}

bool _waitForStatus(anytype hdl, int status, int timeout=5000) {
  int _status = _getLastStatus(hdl);

  while(timeout > 0 && _status != status) {
    _status = _getLastStatus(hdl);

    delay(0,100);
    timeout -= 100;
  }

  return (timeout > 0);
}


/**
 * Mock status CB function, use this to register states.
 */
void mock_statusCB(anytype hdl, int status) {
  if (!mappingHasKey(_asyncStatusMap, hdl)) {
    _asyncStatusMap[hdl] = makeDynInt();
  }

  dynAppend(_asyncStatusMap[hdl], status);
}


//------------------------------------------------------------------------------


bool _getLastMessageFromRb(
  string  rbDp,
  string  &id,
  string  &type,
  string  &proj,
  string  &name,
  anytype &payload,
  bool    test = true
) {
  string val;
  int err = fwRPC_RingBuffer_read(rbDp + ".out", val);

  if (test) {
    assertEqual(0, err, __FUNCTION__ + "Error reading DP: " + rbDp + ".out");
  } else if (err != 0) {
    return false;
  }

  string buffer;
  fwRPC_common_unpackMessage(val, id, type, proj, name, buffer);
  deserializeFromXml(buffer, payload);

  return true;
}


//------------------------------------------------------------------------------


string testUtil_makeStubName(string funcName) {
  return "_" + funcName + "_stub";
}


//------------------------------------------------------------------------------


void _BEGIN_SUITE(string str = "") {
  _TEST_COUNT = 0;

  if (!_PRINT_SUITE) return;
  if (str == "" && !_PRINT_EMPTY) return;
  
  DebugN(
    "--------------------------------------------------------------------------------" +
    "\n# SUITE " + str + "\n" +
    "////////////////////////////////////////////////////////////////////////////////"
  );

  fwRPC_dbg_log_module(
    "TEST",
    fwRPC_DBG_EVENT_TEST,
    "Begin suite " + str
  );

  //force dbg on all modules
  fwRPC_dbg_escalate(fwRPC_DBG_MODULE_CLIENT, 2);
  fwRPC_dbg_escalate(fwRPC_DBG_MODULE_WORKER, 2);
  fwRPC_dbg_escalate(fwRPC_DBG_MODULE_CORE,   2);
  fwRPC_dbg_escalate(fwRPC_DBG_MODULE_BROKER, 2);
}

void _END_SUITE() {
  if (!_PRINT_SUITE) return;

  //stop dbg on all modules
  fwRPC_dbg_deescalate(fwRPC_DBG_MODULE_CLIENT, 0);
  fwRPC_dbg_deescalate(fwRPC_DBG_MODULE_WORKER, 0);
  fwRPC_dbg_deescalate(fwRPC_DBG_MODULE_CORE,   0);
  fwRPC_dbg_deescalate(fwRPC_DBG_MODULE_BROKER, 0);
  
//  DebugN(
//    "\n# END SUITE \n" +
//    "--------------------------------------------------------------------------------"
//  );
}

void _BEGIN_TEST(string str = "") {
  mappingClear(_asyncStatusMap);

  if (str == "") _TEST_COUNT++;

  if (!_PRINT_TEST) return;
  if (str == "" && !_PRINT_EMPTY) return;

  string s = "\n# TEST [" + _TEST_COUNT + "]";
  if (str != "") s += ": " + str;
  s += "\n";

  DebugN(s);

  fwRPC_dbg_log_module(
    "TEST",
    fwRPC_DBG_EVENT_TEST,
    "Begin test " + _TEST_COUNT,
    true
  );
}
